import argparse
from pathlib import Path

from unimport import __version__
from unimport.config import Config

__all__ = (
    "add_sources_option",
    "add_check_option",
    "add_config_option",
    "add_disable_auto_discovery_config_option",
    "add_include_option",
    "add_exclude_option",
    "add_gitignore_option",
    "add_ignore_init_option",
    "add_include_star_import_option",
    "add_diff_option",
    "add_remove_option",
    "add_permission_option",
    "add_version_option",
)


def add_sources_option(parser: argparse.ArgumentParser) -> None:
    parser.add_argument(
        "sources",
        default=Config.default_sources,
        nargs="*",
        help="Files and folders to find the unused imports.",
        action="store",
        type=Path,
    )


def add_check_option(parser: argparse.ArgumentParser) -> None:
    parser.add_argument(
        "--check",
        action="store_true",
        help="Prints which file the unused imports are in.",
        default=Config.check,
    )


def add_config_option(parser: argparse.ArgumentParser) -> None:
    parser.add_argument(
        "-c",
        "--config",
        default=None,
        help="Read configuration from PATH.",
        metavar="PATH",
        action="store",
        type=Path,
    )


def add_disable_auto_discovery_config_option(parser: argparse.ArgumentParser) -> None:
    parser.add_argument(
        "--disable-auto-discovery-config",
        default=Config.disable_auto_discovery_config,
        help=(
            """
            Automatically pick up config options from setup.cfg if it is present in the project
            root else check and if it exists use pyproject.toml.
            """
        ),
        action="store_true",
    )


def add_include_option(parser: argparse.ArgumentParser) -> None:
    parser.add_argument(
        "--include",
        help="File include pattern.",
        metavar="include",
        action="store",
        default=Config.include,
        type=str,
    )


def add_exclude_option(parser: argparse.ArgumentParser) -> None:
    parser.add_argument(
        "--exclude",
        help="File exclude pattern.",
        metavar="exclude",
        action="store",
        default=Config.exclude,
        type=str,
    )


def add_gitignore_option(parser: argparse.ArgumentParser) -> None:
    parser.add_argument(
        "--gitignore",
        action="store_true",
        help="Exclude .gitignore patterns. if present.",
        default=Config.gitignore,
    )


def add_ignore_init_option(parser: argparse.ArgumentParser) -> None:
    parser.add_argument(
        "--ignore-init",
        action="store_true",
        help="Ignore the __init__.py file.",
        default=Config.ignore_init,
    )


def add_include_star_import_option(parser: argparse.ArgumentParser) -> None:
    parser.add_argument(
        "--include-star-import",
        action="store_true",
        help="Include star imports during scanning and refactor.",
        default=Config.include_star_import,
    )


def add_diff_option(parser: argparse.ArgumentParser) -> None:
    parser.add_argument(
        "-d",
        "--diff",
        action="store_true",
        help="Prints a diff of all the changes unimport would make to a file.",
        default=Config.diff,
    )


def add_remove_option(
    exclusive_group: argparse._MutuallyExclusiveGroup,
) -> None:
    exclusive_group.add_argument(
        "-r",
        "--remove",
        action="store_true",
        help="Remove unused imports automatically.",
        default=Config.remove,
    )


def add_permission_option(
    exclusive_group: argparse._MutuallyExclusiveGroup,
) -> None:
    exclusive_group.add_argument(
        "-p",
        "--permission",
        action="store_true",
        help="Refactor permission after see diff.",
        default=Config.permission,
    )


def add_version_option(parser: argparse.ArgumentParser) -> None:
    parser.add_argument(
        "-v",
        "--version",
        action="version",
        version=f"Unimport {__version__}",
        help="Prints version of unimport",
    )


def add_color_option(parser: argparse.ArgumentParser) -> None:
    parser.add_argument(
        "--color",
        default=Config.color,
        type=str,
        metavar="{" + ",".join(Config.get_color_choices()) + "}",
        help="Select whether to use color in the output. Defaults to `%(default)s`.",
    )



from pathlib import Path

from unimport import utils
from unimport.color import difference

__all__ = ("diff",)


def diff(path: Path, source: str, refactor_result: str, use_color: bool = True) -> bool:
    diff_ = utils.diff(source=source, refactor_result=refactor_result, fromfile=path)
    exists_diff = bool(diff_)
    if exists_diff:
        print(difference(diff_, use_color))

    return exists_diff



import contextlib
import dataclasses
from pathlib import Path
from typing import TYPE_CHECKING, Iterator, List, Optional, Sequence, Union

from unimport import commands, utils
from unimport.analyzers import MainAnalyzer
from unimport.color import paint
from unimport.config import Config
from unimport.enums import Color
from unimport.statement import Import

if TYPE_CHECKING:
    from unimport.statement import ImportFrom


__all__ = ("Main",)


@dataclasses.dataclass
class _Result:
    unused_imports: List[Union["Import", "ImportFrom"]] = dataclasses.field(repr=False)
    path: Path
    source: str
    encoding: str
    newline: Optional[str] = None


@dataclasses.dataclass
class Main:
    argv: Optional[Sequence[str]] = None

    config: Config = dataclasses.field(init=False)
    is_syntax_error: bool = dataclasses.field(init=False, default=False)
    is_unused_imports: bool = dataclasses.field(init=False, default=False)
    refactor_applied: bool = dataclasses.field(init=False, default=False)

    def __post_init__(self):
        self.config = self.argv_to_config()

    def argv_to_config(self) -> Config:
        import sys

        from unimport.config import ParseConfig

        return ParseConfig.parse_args(
            commands.generate_parser().parse_args(self.argv if self.argv is not None else sys.argv[1:])
        )

    @contextlib.contextmanager
    def analysis(self, source: str, path: Path) -> Iterator:
        analyzer = MainAnalyzer(
            source=source,
            path=path,
            include_star_import=self.config.include_star_import,
        )
        try:
            analyzer.traverse()
        except SyntaxError as exc:
            print(
                paint(str(exc), Color.RED, self.config.use_color)
                + " at "
                + paint(path.as_posix(), Color.GREEN, self.config.use_color)
            )
            self.is_syntax_error = True

        try:
            yield
        finally:
            analyzer.clear()

    def get_results(self) -> Iterator[_Result]:
        for path in self.config.get_paths():
            source, encoding, newline = utils.read(path)

            with self.analysis(source, path):
                unused_imports = list(Import.get_unused_imports(include_star_import=self.config.include_star_import))
                if self.is_unused_imports is False:
                    self.is_unused_imports = unused_imports != []

                yield _Result(unused_imports, path, source, encoding, newline)

    def check(self, result: _Result) -> None:
        commands.check(result.path, result.unused_imports, self.config.use_color)

    def remove(self, result: _Result, refactor_result):
        commands.remove(
            result.path,
            result.encoding,
            result.newline,
            refactor_result,
            self.config.use_color,
        )
        self.refactor_applied = True

    def diff(self, result, refactor_result):
        return commands.diff(result.path, result.source, refactor_result, self.config.use_color)

    @staticmethod
    def permission(result) -> bool:
        return commands.permission(result.path, result.encoding)

    @classmethod
    def run(cls, argv: Optional[Sequence[str]] = None) -> "Main":
        from unimport.refactor import refactor_string

        self = cls(argv)
        for result in self.get_results():
            if self.config.check:
                self.check(result)
            if any((self.config.diff, self.config.remove)):
                refactor_result = refactor_string(source=result.source, unused_imports=result.unused_imports)
                if self.config.diff:
                    exists_diff = self.diff(result, refactor_result)
                    if self.config.permission and exists_diff:
                        self.config.remove = self.permission(result)
                if self.config.remove and result.source != refactor_result:
                    self.remove(result, refactor_result)
        return self

    def exit_code(self):
        from unimport.utils import return_exit_code

        return return_exit_code(
            is_unused_imports=self.is_unused_imports,
            is_syntax_error=self.is_syntax_error,
            refactor_applied=self.refactor_applied,
        )



import ast

from unimport import constants as C
from unimport import typing as T
from unimport.analyzers.decarators import generic_visit
from unimport.analyzers.utils import first_parent_match, get_parents, set_tree_parents
from unimport.statement import Name, Scope

__all__ = ("NameAnalyzer",)


class NameAnalyzer(ast.NodeVisitor):
    def visit_ClassDef(self, node) -> None:
        Scope.add_current_scope(node)

        self.generic_visit(node)

        Scope.remove_current_scope()

    def visit_FunctionDef(self, node: T.ASTFunctionT) -> None:
        Scope.add_current_scope(node)

        self._type_comment(node)
        self.generic_visit(node)

        Scope.remove_current_scope()

    visit_AsyncFunctionDef = visit_FunctionDef

    def visit_str_helper(self, value: str, node: ast.AST) -> None:
        parent = first_parent_match(node, *C.AST_FUNCTION_TUPLE)
        is_annassign_or_arg = any(isinstance(parent, (ast.AnnAssign, ast.arg)) for parent in get_parents(node))
        if is_annassign_or_arg or (parent is not None and parent.returns is node):
            self.join_visit(value, node)

    def visit_Str(self, node: ast.Str) -> None:
        self.visit_str_helper(node.s, node)

    def visit_Constant(self, node: ast.Constant) -> None:
        if isinstance(node.value, str):
            self.visit_str_helper(node.value, node)

    @generic_visit
    def visit_Name(self, node: ast.Name) -> None:
        if not isinstance(node.parent, ast.Attribute):  # type: ignore
            Name.register(lineno=node.lineno, name=node.id, node=node)

    @generic_visit
    def visit_Attribute(self, node: ast.Attribute) -> None:
        if not isinstance(node.value, ast.Call):
            names = []
            for sub_node in ast.walk(node):
                if isinstance(sub_node, ast.Attribute):
                    names.append(sub_node.attr)
                elif isinstance(sub_node, ast.Name):
                    names.append(sub_node.id)
            names.reverse()
            Name.register(lineno=node.lineno, name=".".join(names), node=node)

    @generic_visit
    def visit_Assign(self, node: ast.Assign) -> None:
        self._type_comment(node)

    @generic_visit
    def visit_arg(self, node: ast.arg) -> None:
        self._type_comment(node)

    @generic_visit
    def visit_Subscript(self, node: ast.Subscript) -> None:
        # type_variable
        # type_var = List["object"] etc.

        def visit_constant_str(node: T.ASTNameType) -> None:
            """Separates the value by node type (str or constant) and gives it
            to the visit function."""

            if isinstance(node, ast.Constant) and isinstance(node.value, str):
                self.join_visit(node.value, node)
            elif isinstance(node, ast.Str):
                self.join_visit(node.s, node)

        if (
            isinstance(node.value, ast.Attribute)
            and isinstance(node.value.value, ast.Name)
            and node.value.value.id == "typing"
        ) or (isinstance(node.value, ast.Name) and node.value.id in C.SUBSCRIPT_TYPE_VARIABLE):
            if C.PY39_PLUS:
                _slice = node.slice
            else:
                _slice = node.slice.value  # type: ignore

            if isinstance(_slice, ast.Tuple):  # type: ignore
                for elt in _slice.elts:  # type: ignore
                    if isinstance(elt, (ast.Constant, ast.Str)):
                        visit_constant_str(elt)
            else:
                if isinstance(_slice, (ast.Constant, ast.Str)):  # type: ignore
                    visit_constant_str(_slice)  # type: ignore

    @generic_visit
    def visit_Call(self, node: ast.Call) -> None:
        # type_variable
        # cast("type", return_value)
        if (
            (
                isinstance(node.func, ast.Attribute)
                and isinstance(node.func.value, ast.Name)
                and node.func.value.id == "typing"
                and node.func.attr == "cast"
            )
            or isinstance(node.func, ast.Name)
            and node.func.id == "cast"
        ):
            if isinstance(node.args[0], ast.Constant) and isinstance(node.args[0].value, str):
                self.join_visit(node.args[0].value, node.args[0])
            elif isinstance(node.args[0], ast.Str):
                self.join_visit(node.args[0].s, node.args[0])

    def _type_comment(self, node: ast.AST) -> None:
        mode = "func_type" if isinstance(node, C.AST_FUNCTION_TUPLE) else "eval"
        type_comment = getattr(node, "type_comment", None)
        if type_comment is not None:
            self.join_visit(type_comment, node, mode=mode)

    def join_visit(self, value: str, node: ast.AST, *, mode: str = "eval") -> None:
        """A function that parses the value, copies locations from the node and
        includes them in self.visit."""
        try:
            tree = ast.parse(value, mode=mode, type_comments=True) if C.PY38_PLUS else ast.parse(value, mode=mode)
        except SyntaxError:
            return None
        else:
            set_tree_parents(tree, parent=node.parent)  # type: ignore
            for new_node in ast.walk(tree):
                ast.copy_location(new_node, node)
            self.visit(tree)

    def traverse(self, tree) -> None:
        self.visit(tree)



import contextlib
import difflib
import functools
import importlib.machinery
import importlib.util
import re
import tokenize
from distutils.util import strtobool
from pathlib import Path
from typing import FrozenSet, Iterable, Iterator, List, Optional, Tuple

from pathspec.patterns.gitwildmatch import GitWildMatchPattern

import unimport.constants as C

__all__ = (
    "get_dir",
    "get_source",
    "get_spec",
    "is_std",
    "action_to_bool",
    "get_exclude_list_from_gitignore",
    "read",
    "list_paths",
    "diff",
    "return_exit_code",
)


@functools.lru_cache(maxsize=128)
def get_dir(package: str) -> FrozenSet[str]:
    try:
        module = importlib.import_module(package)
    except (ImportError, AttributeError, TypeError, ValueError):
        return frozenset()
    return frozenset(dir(module))


@functools.lru_cache(maxsize=128)
def get_source(package: str) -> Optional[str]:
    spec = get_spec(package)
    if spec is not None:
        assert isinstance(spec.loader, importlib.machinery.SourceFileLoader)
        assert isinstance(spec.loader.path, str)
        if spec.loader.path.endswith(".py"):
            return spec.loader.get_data(spec.loader.path).decode("utf-8")

    return None


@functools.lru_cache(maxsize=128)
def get_spec(package: str) -> Optional[importlib.machinery.ModuleSpec]:
    try:
        return importlib.util.find_spec(package)
    except (ImportError, AttributeError, TypeError, ValueError):
        return None


@functools.lru_cache(maxsize=128)
def is_std(package: str) -> bool:
    """Returns True if package module came with from Python."""
    if package in C.BUILTIN_MODULE_NAMES:
        return True
    spec = get_spec(package)
    if spec and isinstance(spec.origin, str):
        return any(
            (
                spec.origin.startswith(C.STDLIB_PATH),
                spec.origin in ("built-in", "frozen"),
                spec.origin.endswith(".so"),
            )
        )

    return False


@functools.lru_cache(maxsize=3)
def action_to_bool(action: str) -> bool:
    if action == "":
        return True
    with contextlib.suppress(ValueError):
        return True if strtobool(action) == 1 else False

    return False


def get_exclude_list_from_gitignore(path=Path(".gitignore")) -> List[GitWildMatchPattern]:
    """Converts .gitignore patterns to regex and return this excludes regex
    list."""

    if not path.is_file():
        return []

    gitignore_patterns: List[GitWildMatchPattern] = []
    source, _, _ = read(path)
    for line in source.splitlines():
        regex, include = GitWildMatchPattern.pattern_to_regex(line)
        if regex:
            pattern = GitWildMatchPattern(re.compile(regex), include)
            gitignore_patterns.append(pattern)

    return gitignore_patterns


def read(path: Path) -> Tuple[str, str, Optional[str]]:
    try:
        with tokenize.open(path) as stream:
            source = stream.read()
            encoding = stream.encoding
            newline = stream.newlines
    except (OSError, SyntaxError):
        return "", "utf-8", None

    # If mixed or unknown newlines, fall back to the platform default
    if not isinstance(newline, str):
        newline = None

    return source, encoding, newline


def list_paths(
    start: Path,
    *,
    include: str = C.INCLUDE_REGEX_PATTERN,
    exclude: str = C.EXCLUDE_REGEX_PATTERN,
    gitignore_patterns: Optional[List[GitWildMatchPattern]] = None,
) -> Iterator[Path]:
    include_regex, exclude_regex = re.compile(include), re.compile(exclude)
    file_names: Iterable[Path] = start.glob(C.GLOB_PATTERN) if start.is_dir() else [start]

    if gitignore_patterns:
        for file_name in file_names:
            if include_regex.search(str(file_name)) and not exclude_regex.search(str(file_name)):
                for gitignore_pattern in gitignore_patterns:
                    match_file = (
                        gitignore_pattern.match_file(str(file_name))
                        if C.PY37_PLUS
                        else list(gitignore_pattern.match([str(file_name)]))
                    )
                    if match_file:
                        break
                else:
                    yield file_name
    else:
        for file_name in file_names:
            if include_regex.search(str(file_name)) and not exclude_regex.search(str(file_name)):
                yield file_name


def diff(*, source: str, refactor_result: str, fromfile: Path = None) -> Tuple[str, ...]:
    return tuple(
        difflib.unified_diff(
            source.splitlines(),
            refactor_result.splitlines(),
            fromfile=fromfile.as_posix() if fromfile else "",
        )
    )


@functools.lru_cache(maxsize=128)
def return_exit_code(*, is_unused_imports: bool, is_syntax_error: bool, refactor_applied: bool) -> int:
    """If this function changes, be sure to update this page
    https://unimport.hakancelik.dev/tutorial/other-useful-features/#exit-code-
    behavior."""

    assert not (
        is_unused_imports is False and refactor_applied is True
    ), "is_unused_imports False while refactor_applied cannot be True."

    if is_syntax_error:
        return 1
    elif is_unused_imports:
        if refactor_applied:
            return 0
        return 1

    return 0



from pathlib import Path

from unimport import utils
from unimport.color import paint
from unimport.enums import Color

__all__ = ("permission",)


def permission(path: Path, use_color: bool) -> bool:
    action = input(f"Apply suggested changes to '{paint(str(path), Color.YELLOW, use_color)}' [Y/n/q] ? >").lower()
    if action == "q":
        raise SystemExit(1)
    return utils.action_to_bool(action)



import ast
from typing import Iterator, Optional, Set

__all__ = ("set_tree_parents", "get_parents", "first_parent_match", "get_defined_names")


def set_tree_parents(tree: ast.AST, parent: Optional[ast.AST] = None) -> None:
    tree.parent = parent  # type: ignore
    for node in ast.walk(tree):
        for child in ast.iter_child_nodes(node):
            child.parent = node  # type: ignore


def get_parents(node: ast.AST) -> Iterator[ast.AST]:
    parent = node
    while parent:
        parent = parent.parent  # type: ignore
        if parent:
            yield parent


def first_parent_match(node: ast.AST, *ancestors):
    try:
        return next(filter(lambda parent: isinstance(parent, ancestors), get_parents(node)))
    except StopIteration:
        return None


def get_defined_names(tree: ast.AST) -> Set[str]:
    defined_names: Set[str] = set()
    for node in ast.walk(tree):
        if isinstance(node, (ast.FunctionDef, ast.AsyncFunctionDef, ast.ClassDef)):
            defined_names.add(node.name)
        elif isinstance(node, ast.Name) and isinstance(node.ctx, ast.Store):
            defined_names.add(node.id)

    return defined_names



import sys

import pytest

from unimport.color import TERMINAL_SUPPORT_COLOR, Color, paint


@pytest.mark.skipif(sys.platform != "win32", reason="Requires Windows")
def test_terminal_support_color_on_win():
    from unimport.color import _enable

    try:
        _enable()
    except OSError:
        assert TERMINAL_SUPPORT_COLOR is False
    else:
        assert TERMINAL_SUPPORT_COLOR is True


@pytest.mark.skipif(sys.platform == "win32", reason="Does not run on Windows")
def test_terminal_support_color():
    assert TERMINAL_SUPPORT_COLOR is True


def test_red_paint():
    text = "test text"

    action_text = paint(text, Color.RED)
    assert Color.RED + text + Color.RESET == action_text


def test_paint_use_color_false():
    text = "test text"

    action_text = paint(text, Color.RED, False)
    assert text == action_text


def test_use_color_setting_true():
    text = "test text"

    action_text = paint(text, Color.RED, True)
    assert Color.RED + text + Color.RESET == action_text



import os

import pytest

import unimport.utils
from tests.utils import reopenable_temp_file
from unimport.main import Main


@pytest.fixture(scope="module")
def source():
    return "\n".join(
        [
            "import os",
            "import sys",
            "",
            "print(sys.executable)\n",
        ]
    )


@pytest.fixture(scope="module")
def result():
    return "\n".join(
        [
            "import sys",
            "",
            "print(sys.executable)\n",
        ]
    )


def test_platform_native_linesep(source, result):
    with reopenable_temp_file(source, newline=os.linesep) as tmp_path:
        Main.run(["--remove", tmp_path.as_posix()])
        with open(tmp_path, encoding="utf-8") as tmp_py_file:
            tmp_py_file.read()
            assert os.linesep == tmp_py_file.newlines
        assert result == unimport.utils.read(tmp_path)[0]


def test_platform_non_native_linesep(non_native_linesep, source, result):
    with reopenable_temp_file(source, newline=non_native_linesep) as tmp_path:
        Main.run(["--remove", tmp_path.as_posix()])
        with open(tmp_path, encoding="utf-8") as tmp_py_file:
            tmp_py_file.read()
            assert non_native_linesep == tmp_py_file.newlines
        assert result == unimport.utils.read(tmp_path)[0]



import argparse
import configparser
import contextlib
import dataclasses
import functools
import sys
from ast import literal_eval
from pathlib import Path
from typing import Any, ClassVar, Dict, Iterator, List, Optional, Tuple

import toml
from pathspec.patterns import GitWildMatchPattern

from unimport import constants as C
from unimport import utils
from unimport.color import TERMINAL_SUPPORT_COLOR

if C.PY38_PLUS:
    from typing import Literal
else:
    from typing_extensions import Literal  # type: ignore

__all__ = ("Config", "ParseConfig")


CONFIG_FILES: Dict[str, str] = {
    "setup.cfg": "unimport",
    "pyproject.toml": "tool.unimport",
}


@dataclasses.dataclass
class Config:
    default_sources: ClassVar[List[Path]] = [Path(".")]  # Not init attribute
    gitignore_patterns: List[GitWildMatchPattern] = dataclasses.field(
        default_factory=list, init=False, repr=False, compare=False
    )  # Not init attribute
    use_color: bool = dataclasses.field(init=False)  # Not init attribute

    sources: Optional[List[Path]] = None
    disable_auto_discovery_config: bool = False
    include: str = C.INCLUDE_REGEX_PATTERN
    exclude: str = C.EXCLUDE_REGEX_PATTERN
    gitignore: bool = False
    remove: bool = False
    diff: bool = False
    include_star_import: bool = False
    permission: bool = False
    check: bool = False
    ignore_init: bool = False
    color: Literal["auto", "always", "never"] = "auto"

    @classmethod
    @functools.lru_cache(maxsize=None)
    def _get_init_fields(cls):
        return [
            key
            for key, field in cls.__dataclass_fields__.items()
            if field._field_type == dataclasses._FIELD and field.init
        ]

    def __post_init__(self):
        if self.sources is None:
            self.sources = self.default_sources

        self.diff = self.diff or self.permission
        self.remove = self.remove or not any((self.diff, self.check))
        self.use_color: bool = self.is_use_color(self.color)

        if self.gitignore:
            self.gitignore_patterns = utils.get_exclude_list_from_gitignore()

        elif self.ignore_init:
            self.exclude = "|".join([self.exclude, C.INIT_FILE_IGNORE_REGEX])

    def get_paths(self) -> Iterator[Path]:
        for source_path in self.sources:
            yield from utils.list_paths(
                source_path,
                include=self.include,
                exclude=self.exclude,
                gitignore_patterns=self.gitignore_patterns,
            )

    @classmethod
    def get_color_choices(cls) -> Tuple[str]:
        return getattr(
            Config.__annotations__["color"],
            "__args__" if C.PY37_PLUS else "__values__",
        )

    @classmethod
    def is_use_color(cls, color: str) -> bool:
        if color not in cls.get_color_choices():
            raise ValueError(color)

        return color == "always" or (color == "auto" and sys.stderr.isatty() and TERMINAL_SUPPORT_COLOR)

    @classmethod
    def build(
        cls,
        *,
        args: Optional[Dict[str, Any]] = None,
        config_context: Optional[Dict[str, Any]] = None,
    ) -> "Config":
        if args is None and config_context is None:
            return cls()

        args = args if args is not None else {}
        config_context = config_context if config_context is not None else {}

        context = {}
        for field_name in cls._get_init_fields():
            config_value = args.get(field_name, None)
            if config_value is None or config_value == getattr(cls, field_name):
                config_value = config_context.get(field_name, getattr(cls, field_name))
            context[field_name] = config_value

        return cls(**context)  # Only init attribute values


@dataclasses.dataclass
class ParseConfig:
    config_file: Path

    def __post_init__(self):
        if not self.config_file.exists():
            raise FileNotFoundError(f"Config file not found: {self.config_file}")

        self.config_section: Optional[str] = CONFIG_FILES.get(self.config_file.name, None)
        if self.config_section is None:
            raise ValueError(f"Unsupported config file: {self.config_file}")

    def parse(self) -> Dict[str, Any]:
        return getattr(self, f"parse_{self.config_file.suffix.strip('.')}")()

    def parse_cfg(self) -> Dict[str, Any]:
        parser = configparser.ConfigParser(allow_no_value=True)
        parser.read(self.config_file)
        if parser.has_section(self.config_section):

            def get_config_as_list(name: str) -> List[str]:
                return literal_eval(
                    parser.get(
                        self.config_section,
                        name,
                        fallback=getattr(Config, name),
                    )
                )

            cfg_context: Dict[str, Any] = {}
            config_annotations_mapping = {
                "sources": List[Path],
                "include": str,
                "exclude": str,
                "gitignore": bool,
                "remove": bool,
                "diff": bool,
                "include_star_import": bool,
                "permission": bool,
                "check": bool,
                "ignore_init": bool,
                "color": str,
            }
            for key, value in parser[self.config_section].items():
                key_type = config_annotations_mapping[key]
                if key_type == bool:
                    cfg_context[key] = parser.getboolean(self.config_section, key)
                elif key_type == str:
                    cfg_context[key] = value  # type: ignore
                elif key_type == List[Path]:
                    cfg_context[key] = [Path(p) for p in get_config_as_list(key)]  # type: ignore
            return cfg_context
        else:
            return {}

    def parse_toml(self) -> Dict[str, Any]:
        parsed_toml = toml.loads(self.config_file.read_text())
        toml_context: Dict[str, Any] = dict(
            functools.reduce(lambda x, y: x.get(y, {}), self.config_section.split("."), parsed_toml)  # type: ignore[attr-defined]
        )
        if toml_context:
            sources = toml_context.get("sources", Config.default_sources)
            toml_context["sources"] = [Path(path) for path in sources]
        return toml_context

    @classmethod
    def parse_args(cls, args: argparse.Namespace) -> "Config":
        config_context: Optional[Dict[str, Any]] = None

        if args.config is not None:
            config_context = cls(args.config).parse()
        elif args.config is None and args.disable_auto_discovery_config is False:
            for path in CONFIG_FILES.keys():
                with contextlib.suppress(FileNotFoundError):
                    config_context = cls(Path(path)).parse()
                    if config_context:
                        break

        if not config_context:
            config_context = None

        return Config.build(args=vars(args), config_context=config_context)



import ast
from typing import List, Optional, Set

from unimport.analyzers.decarators import generic_visit, skip_import
from unimport.analyzers.importable import ImportableAnalyzer
from unimport.statement import Import, ImportFrom, Name, Scope

__all__ = ("ImportAnalyzer",)


class ImportAnalyzer(ast.NodeVisitor):
    __slots__ = (
        "source",
        "include_star_import",
        "any_import_error",
        "defined_names",
    )

    IGNORE_MODULES_IMPORTS = ("__future__",)
    IGNORE_IMPORT_NAMES = ("__all__", "__doc__", "__name__")

    def __init__(
        self, *, source: str, include_star_import: bool = False, defined_names: Optional[Set[str]] = None
    ) -> None:
        self.source = source
        self.include_star_import = include_star_import
        self.defined_names = defined_names or set()

        self.any_import_error = False

        self.if_names: Set[str] = set()
        self.orelse_names: Set[str] = set()

    def traverse(self, tree) -> None:
        self.visit(tree)

    def scope_analysis(self, node):
        Scope.add_current_scope(node)

        self.generic_visit(node)

        Scope.remove_current_scope()

    visit_ClassDef = scope_analysis
    visit_FunctionDef = scope_analysis
    visit_AsyncFunctionDef = scope_analysis

    @generic_visit
    @skip_import
    def visit_Import(self, node: ast.Import) -> None:
        for column, alias in enumerate(node.names):
            name = alias.asname or alias.name
            if name in self.IGNORE_IMPORT_NAMES or (name in self.if_names and name in self.orelse_names):
                continue

            Import.register(lineno=node.lineno, column=column + 1, name=name, package=alias.name, node=node)

    @generic_visit
    @skip_import
    def visit_ImportFrom(self, node: ast.ImportFrom) -> None:
        is_star = node.names[0].name == "*"

        for column, alias in enumerate(node.names):
            package = node.module if not node.level else "." * node.level + str(node.module) or ""
            if (package in self.IGNORE_MODULES_IMPORTS) or (is_star and not self.include_star_import):
                return

            name = package if is_star else (alias.asname or alias.name)
            if name in self.IGNORE_IMPORT_NAMES or (name in self.if_names and name in self.orelse_names):
                continue

            ImportFrom.register(
                lineno=node.lineno,
                column=column + 1,
                name=name,
                package=package,
                star=is_star,
                suggestions=self.get_suggestions(package) if is_star else [],
                node=node,
            )

    def visit_If(self, if_node: ast.If) -> None:
        self.if_names = {
            name.asname or name.name
            for n in filter(lambda node: isinstance(node, (ast.Import, ast.ImportFrom)), if_node.body)
            for name in n.names  # type: ignore
        }

        self.orelse_names = {
            name.asname or name.name
            for n in filter(lambda node: isinstance(node, (ast.Import, ast.ImportFrom)), if_node.orelse)
            for name in n.names  # type: ignore
        }

        self.generic_visit(if_node)

        self.if_names = set()
        self.orelse_names = set()

    def visit_Try(self, node: ast.Try) -> None:
        self.any_import_error = True

        self.generic_visit(node)

        self.any_import_error = False

    def get_suggestions(self, package: str) -> List[str]:
        names = set(map(lambda name: name.name.split(".")[0], Name.names))
        from_names = ImportableAnalyzer.get_names(package)
        return sorted(from_names & (names - self.defined_names))



from pathlib import Path

from unimport.color import paint
from unimport.enums import Color

__all__ = ("remove",)


def remove(path: Path, encoding, newline, refactor_result: str, use_color) -> None:
    with open(path, mode="w", encoding=encoding, newline=newline) as py_file:
        py_file.write(refactor_result)

    print(f"Refactoring '{paint(str(path), Color.GREEN, use_color)}'")



from unimport.commands import options
from unimport.commands.check import check
from unimport.commands.diff import diff
from unimport.commands.parser import generate_parser
from unimport.commands.permission import permission
from unimport.commands.remove import remove

__all__ = (
    "check",
    "diff",
    "permission",
    "remove",
    "options",
    "generate_parser",
)



def main():
    from unimport.color import paint
    from unimport.enums import Color, Emoji
    from unimport.main import Main

    main = Main.run()
    if not main.is_unused_imports and main.config.check:
        print(
            paint(
                f"{Emoji.STAR} Congratulations there is no unused import in your project. {Emoji.STAR}",
                Color.GREEN,
                main.config.use_color,
            )
        )

    raise SystemExit(main.exit_code())


if __name__ == "__main__":
    main()



import ast
from typing import Any, Callable, TypeVar, Union

import libcst as cst

__all__ = (
    "FunctionT",
    "ASTImportableT",
    "ASTFunctionT",
    "ASTNameType",
    "CFNT",
    "CSTImportT",
)

ASTImportableT = Union[ast.AsyncFunctionDef, ast.Attribute, ast.ClassDef, ast.FunctionDef, ast.Name, ast.alias]
ASTFunctionT = TypeVar("ASTFunctionT", ast.FunctionDef, ast.AsyncFunctionDef)
ASTNameType = Union[ast.Name, ast.Constant]

CFNT = TypeVar("CFNT", ast.ClassDef, ast.FunctionDef, ast.AsyncFunctionDef, ast.Name)
CSTImportT = TypeVar("CSTImportT", cst.Import, cst.ImportFrom)

FunctionT = TypeVar("FunctionT", bound=Callable[..., Any])
